﻿Imports Microsoft.VisualBasic
Imports System
Namespace DeepNetTrain
    Friend Class DeepNetTrainProgram
        Shared Sub Main(ByVal args() As String)
            Console.WriteLine(vbCrLf & "Begin deep net training demo " & vbCrLf)

            Dim numInput = 4
            Dim numHidden() As Integer = {10, 10, 10}
            Dim numOutput = 3
            Dim numDataItems = 2000
            Console.WriteLine("Generating " & numDataItems & " artificial training data items ")
            Dim trainData()() As Double = MakeData(numDataItems, numInput, numHidden, numOutput, 5)
            Console.WriteLine(vbCrLf & "Done. Training data is: ")
            ShowMatrix(trainData, 3, 2, True)

            Console.WriteLine(vbCrLf & "Creating a 4-(10,10,10)-3 deep neural network (tanh & softmax) " & vbCrLf)
            Dim dn As New DeepNet(numInput, numHidden, numOutput)

            Dim maxEpochs = 2000
            Dim learnRate As Double = 0.001
            Dim momentum As Double = 0.01
            Console.WriteLine("Setting maxEpochs = " & maxEpochs)
            Console.WriteLine("Setting learnRate = " & learnRate.ToString("F3"))
            Console.WriteLine("Setting momentumm = " & momentum.ToString("F3"))
            Console.WriteLine(vbCrLf & "Start training using back-prop with mean squared error " & vbCrLf)
            Dim wts() As Double = dn.Train(trainData, maxEpochs, learnRate, momentum, 10) ' show error every maxEpochs / 10
            Console.WriteLine("Training complete " & vbCrLf)

            Dim trainError As Double = dn.Error(trainData, False)
            Dim trainAcc As Double = dn.Accuracy(trainData, False)
            Console.WriteLine("Final model MS error = " & trainError.ToString("F4"))
            Console.WriteLine("Final model accuracy = " & trainAcc.ToString("F4"))

            Console.WriteLine(vbCrLf & "End demo ")
            Console.ReadLine()
        End Sub ' Main

        Private Shared Function MakeData(ByVal numItems As Integer, ByVal numInput As Integer, ByVal numHidden() As Integer, ByVal numOutput As Integer, ByVal seed As Integer) As Double()()
            ' generate data using a Deep NN (tanh hidden activation)
            Dim dn As New DeepNet(numInput, numHidden, numOutput) ' make a DNN generator
            Dim rrnd As New Random(seed) ' to make random weights & biases, random input vals
            Dim wtLo As Double = -9.0
            Dim wtHi As Double = 9.0
            Dim nw As Integer = DeepNet.NumWeights(numInput, numHidden, numOutput)
            Dim wts(nw - 1) As Double

            For i = 0 To nw - 1
                wts(i) = (wtHi - wtLo) * rrnd.NextDouble() + wtLo
            Next i
            dn.SetWeights(wts)

            Dim result(numItems - 1)() As Double ' make the result matrix holder
            For r = 0 To numItems - 1
                result(r) = New Double(numInput + numOutput - 1) {} ' allocate the cols
            Next r

            Dim inLo As Double = -4.0 ' pseudo-Gaussian scaling
            Dim inHi As Double = 4.0
            For r = 0 To numItems - 1 ' each row
                Dim inputs(numInput - 1) As Double ' random input values

                For i = 0 To numInput - 1
                    inputs(i) = (inHi - inLo) * rrnd.NextDouble() + inLo
                Next i

                Dim probs() As Double = dn.ComputeOutputs(inputs) ' compute the outputs (as softmax probs) like [0.10, 0.15, 0.55, 0.20]

                Dim outputs() As Double = ProbsToClasses(probs) ' convert to outputs like [0, 0, 1, 0]

                Dim c = 0
                For i = 0 To numInput - 1
                    result(r)(c) = inputs(i)
                    c += 1
                Next i
                For i = 0 To numOutput - 1
                    result(r)(c) = outputs(i)
                    c += 1
                Next i

            Next r
            Return result

        End Function ' MakeData

        Private Shared Function ProbsToClasses(ByVal probs() As Double) As Double()
            Dim result(probs.Length - 1) As Double
            Dim idx = MaxIndex(probs)
            result(idx) = 1.0
            Return result
        End Function

        Private Shared Function MaxIndex(ByVal probs() As Double)
            Dim maxIdx = 0
            Dim maxVal As Double = probs(0)

            For i = 0 To probs.Length - 1
                If probs(i) > maxVal Then
                    maxVal = probs(i)
                    maxIdx = i
                End If
            Next i
            Return maxIdx
        End Function

        Public Shared Sub ShowMatrix(ByVal matrix()() As Double, ByVal numRows As Integer, ByVal decimals As Integer, ByVal indices As Boolean)
            Dim len = matrix.Length.ToString().Length
            For i = 0 To numRows - 1
                If indices = True Then
                    Console.Write("[" & i.ToString().PadLeft(len) & "]  ")
                End If
                For j = 0 To matrix(i).Length - 1
                    Dim v As Double = matrix(i)(j)
                    If v >= 0.0 Then
                        Console.Write(" ") ' '+'
                    End If
                    Console.Write(v.ToString("F" & decimals) & "  ")
                Next j
                Console.WriteLine("")
            Next i

            If numRows < matrix.Length Then
                Console.WriteLine(". . .")
                Dim lastRow = matrix.Length - 1
                If indices = True Then
                    Console.Write("[" & lastRow.ToString().PadLeft(len) & "]  ")
                End If
                For j = 0 To matrix(lastRow).Length - 1
                    Dim v As Double = matrix(lastRow)(j)
                    If v >= 0.0 Then
                        Console.Write(" ") ' '+'
                    End If
                    Console.Write(v.ToString("F" & decimals) & "  ")
                Next j
            End If
            Console.WriteLine(vbCrLf)
        End Sub

        Private Shared Sub ShowMatrix(ByVal matrix()() As Double, ByVal numRows As Integer, ByVal numDec As Integer)
            For r = 0 To numRows - 1
                For c = 0 To matrix(r).Length - 1
                    If matrix(r)(c) >= 0.0 Then ' '+'
                        Console.Write(" ")
                    End If
                    Console.Write(matrix(r)(c).ToString("F" & numDec) & "  ")
                Next c
                Console.WriteLine("")
            Next r
            Console.WriteLine("")
        End Sub

        Private Shared Sub ShowVector(ByVal vector() As Double, ByVal numDec As Integer)
            For i = 0 To vector.Length - 1
                If vector(i) >= 0.0 Then
                    Console.Write(" ")
                End If
                Console.Write(vector(i).ToString("F" & numDec) & "  ")
            Next i
            Console.WriteLine("")
        End Sub

    End Class ' Program

    Public Class DeepNet
        Public Shared rnd As Random ' weight init and train shuffle

        Public nInput   ' number input nodes
        Public nHidden() As Integer ' number hidden nodes, each layer
        Public nOutput   ' number output nodes
        Public nLayers   ' number hidden node layers

        Public iNodes() As Double ' input nodes
        Public hNodes()() As Double
        Public oNodes() As Double

        Public ihWeights()() As Double ' input- 1st hidden
        Public hhWeights()()() As Double ' hidden-hidden
        Public hoWeights()() As Double ' last hidden-output

        Public hBiases()() As Double ' hidden node biases
        Public oBiases() As Double ' output node biases

        Public ihGradient00 As Double ' one gradient to monitor

        Public Sub New(ByVal numInput As Integer, ByVal numHidden() As Integer, ByVal numOutput As Integer)
            rnd = New Random(0) ' seed could be a ctor parameter

            Me.nInput = numInput
            Me.nHidden = New Integer(numHidden.Length - 1) {}
            For i = 0 To numHidden.Length - 1
                Me.nHidden(i) = numHidden(i)
            Next i
            Me.nOutput = numOutput
            Me.nLayers = numHidden.Length

            iNodes = New Double(numInput - 1) {}
            hNodes = MakeJaggedMatrix(numHidden)
            oNodes = New Double(numOutput - 1) {}

            ihWeights = MakeMatrix(numInput, numHidden(0))
            hoWeights = MakeMatrix(numHidden(nLayers - 1), numOutput)

            hhWeights = New Double(nLayers - 2)()() {} ' if 3 h layer, 2 h-h weights[][]
            For h = 0 To hhWeights.Length - 1
                Dim rows = numHidden(h)
                Dim cols = numHidden(h + 1)
                hhWeights(h) = MakeMatrix(rows, cols)
            Next h

            hBiases = MakeJaggedMatrix(numHidden) ' pass an array of lengths
            oBiases = New Double(numOutput - 1) {}

            InitializeWeights() ' small randomm non-zero values
        End Sub ' ctor

        Public Sub InitializeWeights()
            ' make wts
            Dim lo As Double = -0.1
            Dim hi As Double = +0.1
            Dim numWts = DeepNet.NumWeights(Me.nInput, Me.nHidden, Me.nOutput)
            Dim wts(numWts - 1) As Double
            For i = 0 To numWts - 1
                wts(i) = (hi - lo) * rnd.NextDouble() + lo
            Next i
            Me.SetWeights(wts)
        End Sub

        Public Function ComputeOutputs(ByVal xValues() As Double) As Double()
            ' 'xValues' might have class label or not
            ' copy vals into iNodes
            For i = 0 To nInput - 1 ' possible trunc
                iNodes(i) = xValues(i)
            Next i

            ' zero-out all hNodes, oNodes
            For h = 0 To nLayers - 1
                For j = 0 To nHidden(h) - 1
                    hNodes(h)(j) = 0.0
                Next j
            Next h

            For k = 0 To nOutput - 1
                oNodes(k) = 0.0
            Next k

            ' input to 1st hid layer
            For j = 0 To nHidden(0) - 1 ' each hidden node, 1st layer
                For i = 0 To nInput - 1
                    hNodes(0)(j) += ihWeights(i)(j) * iNodes(i)
                Next i
                ' add the bias
                hNodes(0)(j) += hBiases(0)(j)
                ' apply activation
                hNodes(0)(j) = Math.Tanh(hNodes(0)(j))
            Next j

            ' each remaining hidden node
            For h = 1 To nLayers - 1
                For j = 0 To nHidden(h) - 1 ' 'to index'
                    For jj = 0 To nHidden(h - 1) - 1 ' 'from index'
                        hNodes(h)(j) += hhWeights(h - 1)(jj)(j) * hNodes(h - 1)(jj)
                    Next jj
                    hNodes(h)(j) += hBiases(h)(j) ' add bias value
                    hNodes(h)(j) = Math.Tanh(hNodes(h)(j)) ' apply activation
                Next j
            Next h

            ' compute ouput node values
            For k = 0 To nOutput - 1
                For j = 0 To nHidden(nLayers - 1) - 1
                    oNodes(k) += hoWeights(j)(k) * hNodes(nLayers - 1)(j)
                Next j
                oNodes(k) += oBiases(k) ' add bias value

            Next k

            Dim retResult() As Double = Softmax(oNodes) ' softmax activation all oNodes

            For k = 0 To nOutput - 1
                oNodes(k) = retResult(k)
            Next k
            Return retResult ' calling convenience

        End Function ' ComputeOutputs

        Public Function Train(ByVal trainData()() As Double, ByVal maxEpochs As Integer, ByVal learnRate As Double, ByVal momentum As Double, ByVal showEvery As Integer) As Double()
            ' no momentum right now
            ' each weight (and bias) needs a big_delta. big_delta is just learnRate * "a gradient"
            ' so goal is to find "a gradient".
            ' the gradient (the term can have several meanings) is "a signal" * "an input"
            ' the signal 

            ' 1. each weight and bias has a 'gradient' (partial dervative)
            Dim hoGrads()() As Double = MakeMatrix(nHidden(nLayers - 1), nOutput) ' last_hidden layer - output weights grads
            Dim hhGrads(nLayers - 2)()() As Double
            For h As Integer = 0 To hhGrads.Length - 1
                Dim rows As Integer = nHidden(h)
                Dim cols = nHidden(h + 1)
                hhGrads(h) = MakeMatrix(rows, cols)
            Next h
            Dim ihGrads()() As Double = MakeMatrix(nInput, nHidden(0)) ' input-first_hidden wts gradients
            ' biases
            Dim obGrads(nOutput - 1) As Double ' output node bias grads
            Dim hbGrads()() As Double = MakeJaggedMatrix(nHidden) ' hidden node bias grads

            ' 2. each output node and each hidden node has a 'signal' == gradient without associated input (lower case delta in Wikipedia)
            Dim oSignals(nOutput - 1) As Double
            Dim hSignals()() As Double = MakeJaggedMatrix(nHidden)

            ' 3. for momentum, each weight and bias needs to store the prev epoch delta
            ' the structure for prev deltas is same as for Weights & Biases, which is same as for Grads

            Dim hoPrevWeightsDelta()() As Double = MakeMatrix(nHidden(nLayers - 1), nOutput) ' last_hidden layer - output weights momentum term
            Dim hhPrevWeightsDelta(nLayers - 2)()() As Double
            For h = 0 To hhPrevWeightsDelta.Length - 1
                Dim rows = nHidden(h)
                Dim cols = nHidden(h + 1)
                hhPrevWeightsDelta(h) = MakeMatrix(rows, cols)
            Next h
            Dim ihPrevWeightsDelta()() As Double = MakeMatrix(nInput, nHidden(0)) ' input-first_hidden wts gradients
            Dim oPrevBiasesDelta(nOutput - 1) As Double ' output node bias prev deltas
            Dim hPrevBiasesDelta()() As Double = MakeJaggedMatrix(nHidden) ' hidden node bias prev deltas

            Dim epoch = 0
            Dim xValues(nInput - 1) As Double ' not necessary - could copy direct from data item to iNodes
            Dim tValues(nOutput - 1) As Double ' not necessary
            Dim derivative As Double = 0.0 ' of activation (softmax or tanh or log-sigmoid or relu)
            Dim errorSignal As Double = 0.0 ' target - output

            Dim sequence(trainData.Length - 1) As Integer
            For i As Integer = 0 To sequence.Length - 1
                sequence(i) = i
            Next i

            Dim errInterval As Integer = maxEpochs \ showEvery ' interval to check & display error
            Do While epoch < maxEpochs
                epoch += 1
                If epoch Mod errInterval = 0 AndAlso epoch < maxEpochs Then ' display curr MSE
                    Dim trainErr As Double = [Error](trainData, False) ' using curr weights & biases
                    Dim trainAcc As Double = Me.Accuracy(trainData, False)
                    Console.Write("epoch = " & epoch & "  MS error = " & trainErr.ToString("F4"))
                    Console.WriteLine("  accuracy = " & trainAcc.ToString("F4"))

                    Console.WriteLine("input-to-hidden [0][0] gradient = " & Me.ihGradient00.ToString("F6"))
                    Console.WriteLine("")

                End If

                Shuffle(sequence) ' must visit each training data in random order in stochastic GD

                For ii = 0 To trainData.Length - 1 ' each train data item
                    Dim idx = sequence(ii) ' idx points to a data item
                    Array.Copy(trainData(idx), xValues, nInput)
                    Array.Copy(trainData(idx), nInput, tValues, 0, nOutput)
                    ComputeOutputs(xValues) ' copy xValues in, compute outputs using curr weights & biases, ignore return

                    ' must compute signals from right-to-left
                    ' weights and bias gradients can be computed left-to-right
                    ' weights and bias gradients can be updated left-to-right

                    ' x. compute output node signals (assumes softmax) depends on target values to the right
                    For k = 0 To nOutput - 1
                        errorSignal = tValues(k) - oNodes(k) ' Wikipedia uses (o-t)
                        derivative = (1 - oNodes(k)) * oNodes(k) ' for softmax (same as log-sigmoid) with MSE
                        'derivative = 1.0;  // for softmax with cross-entropy
                        oSignals(k) = errorSignal * derivative ' we'll use this for ho-gradient and hSignals
                    Next k

                    ' x. compute signals for last hidden layer (depends on oNodes values to the right)
                    Dim lastLayer = nLayers - 1
                    For j = 0 To nHidden(lastLayer) - 1
                        derivative = (1 + hNodes(lastLayer)(j)) * (1 - hNodes(lastLayer)(j)) ' for tanh!
                        Dim sum As Double = 0.0 ' need sums of output signals times hidden-to-output weights
                        For k = 0 To nOutput - 1
                            sum += oSignals(k) * hoWeights(j)(k) ' represents error signal
                        Next k
                        hSignals(lastLayer)(j) = derivative * sum
                    Next j

                    ' x. compute signals for all the non-last layer hidden nodes (depends on layer to the right)
                    For h = lastLayer - 1 To 0 Step -1 ' each hidden layer, right-to-left
                        For j = 0 To nHidden(h) - 1 ' each node
                            derivative = (1 + hNodes(h)(j)) * (1 - hNodes(h)(j)) ' for tanh

                            Dim sum As Double = 0.0 ' need sums of output signals times hidden-to-output weights

                            For jj = 0 To nHidden(h + 1) - 1 ' layer to right of curr layer
                                sum += hSignals(h + 1)(jj) * hhWeights(h)(j)(jj)
                            Next jj

                            hSignals(h)(j) = derivative * sum

                        Next j ' j
                    Next h ' h

                    ' at this point, all hidden and output node signals have been computed
                    ' calculate gradients left-to-right

                    ' x. compute input-to-hidden weights gradients using iNodes & hSignal[0]
                    For i = 0 To nInput - 1
                        For j = 0 To nHidden(0) - 1
                            ihGrads(i)(j) = iNodes(i) * hSignals(0)(j) ' "from" input & "to" signal
                        Next j
                    Next i

                    ' save the special monitored ihGradient00
                    Me.ihGradient00 = ihGrads(0)(0)

                    ' x. compute hidden-to-hidden gradients
                    For h = 0 To nLayers - 2
                        For j = 0 To nHidden(h) - 1
                            For jj = 0 To nHidden(h + 1) - 1
                                hhGrads(h)(j)(jj) = hNodes(h)(j) * hSignals(h + 1)(jj)
                            Next jj
                        Next j
                    Next h

                    ' x. compute hidden-to-output gradients
                    For j = 0 To nHidden(lastLayer) - 1
                        For k = 0 To nOutput - 1
                            hoGrads(j)(k) = hNodes(lastLayer)(j) * oSignals(k) ' from last hidden, to oSignals
                        Next k
                    Next j

                    ' compute bias gradients
                    ' a bias is like a weight on the left/before
                    ' so there's a dummy input of 1.0 and we use the signal of the 'current' layer

                    ' x. compute all hidden bias gradients
                    ' a gradient needs the "left/from" input and the "right/to" signal
                    ' for biases we use a dummy 1.0 input

                    For h = 0 To nLayers - 1
                        For j = 0 To nHidden(h) - 1
                            hbGrads(h)(j) = 1.0 * hSignals(h)(j)
                        Next j
                    Next h

                    ' x. output bias gradients
                    For k = 0 To nOutput - 1
                        obGrads(k) = 1.0 * oSignals(k)
                    Next k

                    ' at this point all signals have been computed and all gradients have been computed 
                    ' so can use gradients to update all weights and biases.
                    ' save each delta for the momentum

                    ' x. update input-to-first_hidden weights using ihWeights & ihGrads
                    For i = 0 To nInput - 1
                        For j = 0 To nHidden(0) - 1
                            Dim delta As Double = ihGrads(i)(j) * learnRate
                            ihWeights(i)(j) += delta
                            ihWeights(i)(j) += ihPrevWeightsDelta(i)(j) * momentum
                            ihPrevWeightsDelta(i)(j) = delta
                        Next j
                    Next i

                    ' other hidden-to-hidden weights using hhWeights & hhGrads
                    For h = 0 To nLayers - 2
                        For j = 0 To nHidden(h) - 1
                            For jj = 0 To nHidden(h + 1) - 1
                                Dim delta As Double = hhGrads(h)(j)(jj) * learnRate
                                hhWeights(h)(j)(jj) += delta
                                hhWeights(h)(j)(jj) += hhPrevWeightsDelta(h)(j)(jj) * momentum
                                hhPrevWeightsDelta(h)(j)(jj) = delta
                            Next jj
                        Next j
                    Next h

                    ' update hidden-to-output weights using hoWeights & hoGrads
                    For j = 0 To nHidden(lastLayer) - 1
                        For k = 0 To nOutput - 1
                            Dim delta As Double = hoGrads(j)(k) * learnRate
                            hoWeights(j)(k) += delta
                            hoWeights(j)(k) += hoPrevWeightsDelta(j)(k) * momentum
                            hoPrevWeightsDelta(j)(k) = delta
                        Next k
                    Next j

                    ' update hidden biases using hBiases & hbGrads
                    For h = 0 To nLayers - 1
                        For j = 0 To nHidden(h) - 1
                            Dim delta As Double = hbGrads(h)(j) * learnRate
                            hBiases(h)(j) += delta
                            hBiases(h)(j) += hPrevBiasesDelta(h)(j) * momentum
                            hPrevBiasesDelta(h)(j) = delta
                        Next j
                    Next h

                    ' update output biases using oBiases & obGrads
                    For k = 0 To nOutput - 1
                        Dim delta As Double = obGrads(k) * learnRate
                        oBiases(k) += delta
                        oBiases(k) += oPrevBiasesDelta(k) * momentum
                        oPrevBiasesDelta(k) = delta
                    Next k

                    ' Whew!
                Next ii ' for each train data item
            Loop ' while

            Dim bestWts() As Double = Me.GetWeights
            Return bestWts
        End Function ' Train

        Public Function [Error](ByVal data()() As Double, ByVal verbose As Boolean) As Double
            ' mean squared error using current weights & biases
            Dim sumSquaredError As Double = 0.0
            Dim xValues(nInput - 1) As Double ' first numInput values in trainData
            Dim tValues(nOutput - 1) As Double ' last numOutput values

            ' walk thru each training case. looks like (6.9 3.2 5.7 2.3) (0 0 1)
            For i = 0 To data.Length - 1
                Array.Copy(data(i), xValues, nInput)
                Array.Copy(data(i), nInput, tValues, 0, nOutput) ' get target values
                Dim yValues() As Double = Me.ComputeOutputs(xValues) ' outputs using current weights

                If verbose = True Then
                    ShowVector(yValues, 4)
                    ShowVector(tValues, 4)
                    Console.WriteLine("")
                    Console.ReadLine()
                End If

                For j = 0 To nOutput - 1
                    Dim err As Double = tValues(j) - yValues(j)
                    sumSquaredError += err * err
                Next j
            Next i
            Return sumSquaredError / (data.Length * nOutput) ' average per item
        End Function ' Error

        Public Function [Error](ByVal data()() As Double, ByVal weights() As Double) As Double
            ' mean squared error using supplied weights & biases
            Me.SetWeights(weights)

            Dim sumSquaredError As Double = 0.0
            Dim xValues(nInput - 1) As Double ' first numInput values in trainData
            Dim tValues(nOutput - 1) As Double ' last numOutput values

            ' walk thru each training case. looks like (6.9 3.2 5.7 2.3) (0 0 1)
            For i = 0 To data.Length - 1
                Array.Copy(data(i), xValues, nInput)
                Array.Copy(data(i), nInput, tValues, 0, nOutput) ' get target values
                Dim yValues() As Double = Me.ComputeOutputs(xValues) ' outputs using current weights
                For j = 0 To nOutput - 1
                    Dim err As Double = tValues(j) - yValues(j)
                    sumSquaredError += err * err
                Next j
            Next i
            Return sumSquaredError / data.Length
        End Function ' Error

        Public Function Accuracy(ByVal data()() As Double, ByVal verbose As Boolean) As Double
            ' percentage correct using winner-takes all
            Dim numCorrect = 0
            Dim numWrong = 0
            Dim xValues(nInput - 1) As Double ' inputs
            Dim tValues(nOutput - 1) As Double ' targets
            Dim yValues() As Double ' computed Y

            For i = 0 To data.Length - 1
                Array.Copy(data(i), xValues, nInput) ' get x-values
                Array.Copy(data(i), nInput, tValues, 0, nOutput) ' get t-values
                yValues = Me.ComputeOutputs(xValues)

                If verbose = True Then
                    ShowVector(yValues, 4)
                    ShowVector(tValues, 4)
                    Console.WriteLine("")
                    Console.ReadLine()
                End If

                Dim maxIndex = DeepNet.MaxIndex(yValues) ' which cell in yValues has largest value?
                Dim tMaxIndex = DeepNet.MaxIndex(tValues)

                If maxIndex = tMaxIndex Then
                    numCorrect += 1
                Else
                    numWrong += 1
                End If
            Next i
            Return (numCorrect * 1.0) / (numCorrect + numWrong)
        End Function

        Private Shared Sub ShowVector(ByVal vector() As Double, ByVal dec As Integer)
            For i = 0 To vector.Length - 1
                Console.Write(vector(i).ToString("F" & dec) & " ")
            Next i
            Console.WriteLine("")
        End Sub

        Public Function Accuracy(ByVal data()() As Double, ByVal weights() As Double) As Double
            Me.SetWeights(weights)
            ' percentage correct using winner-takes all
            Dim numCorrect = 0
            Dim numWrong = 0
            Dim xValues(nInput - 1) As Double ' inputs
            Dim tValues(nOutput - 1) As Double ' targets
            Dim yValues() As Double ' computed Y

            For i = 0 To data.Length - 1
                Array.Copy(data(i), xValues, nInput) ' get x-values
                Array.Copy(data(i), nInput, tValues, 0, nOutput) ' get t-values
                yValues = Me.ComputeOutputs(xValues)
                Dim maxIndex = DeepNet.MaxIndex(yValues) ' which cell in yValues has largest value?
                Dim tMaxIndex = DeepNet.MaxIndex(tValues)

                If maxIndex = tMaxIndex Then
                    numCorrect += 1
                Else
                    numWrong += 1
                End If
            Next i
            Return (numCorrect * 1.0) / (numCorrect + numWrong)
        End Function

        Private Shared Function MaxIndex(ByVal vector() As Double)   ' helper for Accuracy()
            ' index of largest value in vector[]
            Dim bigIndex = 0
            Dim biggestVal As Double = vector(0)
            For i = 0 To vector.Length - 1
                If vector(i) > biggestVal Then
                    biggestVal = vector(i)
                    bigIndex = i
                End If
            Next i
            Return bigIndex
        End Function

        Private Sub Shuffle(ByVal sequence() As Integer) ' instance method
            For i = 0 To sequence.Length - 1
                Dim r = rnd.Next(i, sequence.Length)
                Dim tmp = sequence(r)
                sequence(r) = sequence(i)
                sequence(i) = tmp
            Next i
        End Sub ' Shuffle

        Private Shared Function MyTanh(ByVal x As Double) As Double
            If x < -20.0 Then ' approximation is correct to 30 decimals
                Return -1.0
            ElseIf x > 20.0 Then
                Return 1.0
            Else
                Return Math.Tanh(x)
            End If
        End Function


        Private Shared Function Softmax(ByVal oSums() As Double) As Double()
            ' does all output nodes at once so scale
            ' doesn't have to be re-computed each time.
            ' possible overflow . . . use max trick

            Dim sum As Double = 0.0
            For i = 0 To oSums.Length - 1
                sum += Math.Exp(oSums(i))
            Next i

            Dim result(oSums.Length - 1) As Double
            For i = 0 To oSums.Length - 1
                result(i) = Math.Exp(oSums(i)) / sum
            Next i

            Return result ' now scaled so that xi sum to 1.0
        End Function


        Public Sub SetWeights(ByVal wts() As Double)
            ' order: ihweights - hhWeights[] - hoWeights - hBiases[] - oBiases
            Dim nw = NumWeights(Me.nInput, Me.nHidden, Me.nOutput) ' total num wts + biases
            If wts.Length <> nw Then
                Throw New Exception("Bad wts[] length in SetWeights()")
            End If
            Dim ptr = 0 ' pointer into wts[]

            For i = 0 To nInput - 1 ' input node
                For j = 0 To hNodes(0).Length - 1 ' 1st hidden layer nodes
                    ihWeights(i)(j) = wts(ptr)
                    ptr += 1
                Next j
            Next i

            For h = 0 To nLayers - 2 ' not last h layer
                For j = 0 To nHidden(h) - 1 ' from node
                    For jj = 0 To nHidden(h + 1) - 1 ' to node
                        hhWeights(h)(j)(jj) = wts(ptr)
                        ptr += 1
                    Next jj
                Next j
            Next h

            Dim hi = Me.nLayers - 1 ' if 3 hidden layers (0,1,2) last is 3-1 = [2]
            For j = 0 To Me.nHidden(hi) - 1
                For k = 0 To Me.nOutput - 1
                    hoWeights(j)(k) = wts(ptr)
                    ptr += 1
                Next k
            Next j

            For h = 0 To nLayers - 1 ' hidden node biases
                For j = 0 To Me.nHidden(h) - 1
                    hBiases(h)(j) = wts(ptr)
                    ptr += 1
                Next j
            Next h

            For k = 0 To nOutput - 1
                oBiases(k) = wts(ptr)
                ptr += 1
            Next k
        End Sub ' SetWeights

        Public Function GetWeights() As Double()
            ' order: ihweights -> hhWeights[] -> hoWeights -> hBiases[] -> oBiases
            Dim nw = NumWeights(Me.nInput, Me.nHidden, Me.nOutput) ' total num wts + biases
            Dim result(nw - 1) As Double

            Dim ptr = 0 ' pointer into result[]

            For i = 0 To nInput - 1 ' input node
                For j = 0 To hNodes(0).Length - 1 ' 1st hidden layer nodes
                    result(ptr) = ihWeights(i)(j)
                    ptr += 1
                Next j
            Next i

            For h = 0 To nLayers - 2 ' not last h layer
                For j = 0 To nHidden(h) - 1 ' from node
                    For jj = 0 To nHidden(h + 1) - 1 ' to node
                        result(ptr) = hhWeights(h)(j)(jj)
                        ptr += 1
                    Next jj
                Next j
            Next h

            Dim hi = Me.nLayers - 1 ' if 3 hidden layers (0,1,2) last is 3-1 = [2]
            For j = 0 To Me.nHidden(hi) - 1
                For k = 0 To Me.nOutput - 1
                    result(ptr) = hoWeights(j)(k)
                    ptr += 1
                Next k
            Next j

            For h = 0 To nLayers - 1 ' hidden node biases
                For j = 0 To Me.nHidden(h) - 1
                    result(ptr) = hBiases(h)(j)
                    ptr += 1
                Next j
            Next h

            For k = 0 To nOutput - 1
                result(ptr) = oBiases(k)
                ptr += 1
            Next k
            Return result
        End Function

        Public Shared Function NumWeights(ByVal numInput As Integer, ByVal numHidden() As Integer, ByVal numOutput As Integer) As Integer
            ' total num weights and biases
            Dim ihWts As Integer = numInput * numHidden(0)

            Dim hhWts = 0
            For j = 0 To numHidden.Length - 2
                Dim rows = numHidden(j)
                Dim cols = numHidden(j + 1)
                hhWts += rows * cols
            Next j
            Dim hoWts = numHidden(numHidden.Length - 1) * numOutput

            Dim hbs = 0
            For i = 0 To numHidden.Length - 1
                hbs += numHidden(i)
            Next i

            Dim obs = numOutput
            Dim nw = ihWts + hhWts + hoWts + hbs + obs
            Return nw
        End Function

        Public Shared Function MakeJaggedMatrix(ByVal cols() As Integer) As Double()()
            ' array of arrays using size info in cols[]
            Dim rows As Integer = cols.Length ' num rows inferred by col count
            Dim result(rows - 1)() As Double
            For i = 0 To rows - 1
                Dim ncol = cols(i)
                result(i) = New Double(ncol - 1) {}
            Next i
            Return result
        End Function

        Public Shared Function MakeMatrix(ByVal rows As Integer, ByVal cols As Integer) As Double()()
            Dim result(rows - 1)() As Double
            For i = 0 To rows - 1
                result(i) = New Double(cols - 1) {}
            Next i
            Return result
        End Function

        Public Sub Dump()
            For i = 0 To nInput - 1
                Console.WriteLine("input node [" & i & "] = " & iNodes(i).ToString("F4"))
            Next i
            For h = 0 To nLayers - 1
                Console.WriteLine("")
                For j = 0 To nHidden(h) - 1
                    Console.WriteLine("hidden layer " & h & " node [" & j & "] = " & hNodes(h)(j).ToString("F4"))
                Next j
            Next h
            Console.WriteLine("")
            For k = 0 To nOutput - 1
                Console.WriteLine("output node [" & k & "] = " & oNodes(k).ToString("F4"))
            Next k

            Console.WriteLine("")
            For i = 0 To nInput - 1
                For j = 0 To nHidden(0) - 1
                    Console.WriteLine("input-hidden wt [" & i & "][" & j & "] = " & ihWeights(i)(j).ToString("F4"))
                Next j
            Next i

            For h = 0 To nLayers - 2 ' note
                Console.WriteLine("")
                For j = 0 To nHidden(h) - 1
                    For jj = 0 To nHidden(h + 1) - 1
                        Console.WriteLine("hidden-hidden wt layer " & h & " to layer " & (h + 1) & " node [" & j & "] to [" & jj & "] = " & hhWeights(h)(j)(jj).ToString("F4"))
                    Next jj
                Next j
            Next h

            Console.WriteLine("")
            For j = 0 To nHidden(nLayers - 1) - 1
                For k = 0 To nOutput - 1
                    Console.WriteLine("hidden-output wt [" & j & "][" & k & "] = " & hoWeights(j)(k).ToString("F4"))
                Next k
            Next j

            For h = 0 To nLayers - 1
                Console.WriteLine("")
                For j = 0 To nHidden(h) - 1
                    Console.WriteLine("hidden layer " & h & " bias [" & j & "] = " & hBiases(h)(j).ToString("F4"))
                Next j
            Next h

            Console.WriteLine("")
            For k = 0 To nOutput - 1
                Console.WriteLine("output node bias [" & k & "] = " & oBiases(k).ToString("F4"))
            Next k

        End Sub ' Dump

    End Class ' class DeepNet

End Namespace ' ns